"""
Utility functions
-----------------
Functions :func:`complex_to_seq` and :func:`seq_to_complex` 
are useful to convert between the matrix representation of 
complex numbers and Python :obj:`complex`.

The function :func:`mean` evaluates the mean of a sequence.

The function :func:`implicit` will evaluate the solution 
to :math:`fn(x) = 0`

Module contents
---------------

"""
from __future__ import division

import math
import numpy as np 

try:  # Python 2
    import __builtin__ as builtins
    from collections import Iterable
    from itertools import izip
except ImportError:
    import builtins
    from collections.abc import Iterable
    izip = zip
    xrange = range

from GTC import (
    is_sequence,
    inf ,
    EPSILON
)

from GTC.lib import (
    UncertainReal, UncertainComplex, 
    mult_2nd_real_pair, mult_2nd_complex_pair,mult_2nd_real_complex
)

from GTC import vector
        
__all__ = (
    'complex_to_seq',
    'seq_to_complex',
    'mean',
    'implicit'
)
    
#---------------------------------------------------------------------------
def sum(seq,*args,**kwargs):
    """Return the sum of elements in `seq`
    
    :arg seq: a sequence, :class:`~numpy.ndarray`, or iterable, of numbers or uncertain numbers
    :arg args: optional arguments when ``seq`` is an :class:`~numpy.ndarray`
    :arg kwargs: optional keyword arguments when ``seq`` is an :class:`~numpy.ndarray`
    
    .. versionadded:: 1.1

    """
    if isinstance(seq,np.ndarray):
        return np.asarray(seq).sum(*args, **kwargs)
        
    elif is_sequence(seq) or isinstance(seq,Iterable):
        return builtins.sum(seq)
        
    else:
        raise RuntimeError(
            "{!r} is not iterable".format(seq)
        )    
 
#---------------------------------------------------------------------------
def mean(seq,*args,**kwargs):
    """Return the arithmetic mean of the elements in `seq`
    
    :arg seq: a sequence, :class:`~numpy.ndarray`, or iterable, of numbers or uncertain numbers
    :arg args: optional arguments when ``seq`` is an :class:`~numpy.ndarray`
    :arg kwargs: optional keyword arguments when ``seq`` is an :class:`~numpy.ndarray`
    
    If the elements of ``seq`` are uncertain numbers, 
    an uncertain number is returned.
    
    **Example** ::
    
        >>> seq = [ ureal(1,1), ureal(2,1), ureal(3,1) ]
        >>> function.mean(seq)
        ureal(2.0,0.5773502691896257,inf)
        
    """
    if is_sequence(seq):
        return sum(seq)/len(seq)
        
    elif isinstance(seq,np.ndarray):
        return np.asarray(seq).mean(*args, **kwargs)
        
    elif isinstance(seq,Iterable):
        count = 0
        total = 0
        for i in seq:
            total += i
            count += 1
        return total/count
        
    else:
        raise RuntimeError(
            "{!r} is not iterable".format(seq)
        )
#---------------------------------------------------------------------------
def complex_to_seq(z):
    """Transform a complex number into a 4-element sequence

    :arg z: a number

    If ``z = x + yj``, then an array of the form ``[[x,-y],[y,x]]`` 
    can be used to represent ``z`` in matrix computations. 

    **Examples**::
        >>> import numpy
        >>> z = 1 + 2j
        >>> function.complex_to_seq(z)
        (1.0, -2.0, 2.0, 1.0)
        
        >>> m = numpy.array( function.complex_to_seq(z) )
        >>> m.shape = (2,2)
        >>> print( m )
        [[ 1. -2.]
         [ 2.  1.]]
        
    """
    z = complex(z)
    return (z.real,-z.imag,z.imag,z.real)

#---------------------------------------------------------------------------
def seq_to_complex(seq):
    """Transform a 4-element sequence into a complex number 

    :arg seq:   a 4-element sequence
    :raises RuntimeError: if ``seq`` is ill-conditioned

    **Examples**::

        >>> import numpy
        >>> seq = (1,-2,2,1)
        >>> z = function.seq_to_complex( seq )
        >>> z 
        (1+2j)
        >>> a = numpy.array((1,-2,2,1))
        >>> a.shape = 2,2
        >>> a
        array([[ 1, -2],
               [ 2,  1]])
        >>> z = function.seq_to_complex(a)
        >>> z 
        (1+2j)

    """
    if hasattr(seq,'shape'):
        if seq.shape != (2,2):
            raise RuntimeError("array shape illegal: {}".format(seq))
        elif (
            math.fabs( seq[0,0] - seq[1,1] ) > EPSILON
        or  math.fabs( seq[1,0] + seq[0,1] ) > EPSILON ):
            raise RuntimeError("ill-conditioned sequence: {}".format(seq))
        else:
            seq = list( seq.flat )
            
    elif is_sequence(seq):
        if len(seq) != 4:
            raise RuntimeError("sequence must have 4 elements: {}".format(seq))
        elif (
            math.fabs( seq[0] - seq[3] ) > EPSILON
        or  math.fabs( seq[1] + seq[2] ) > EPSILON ):
            raise RuntimeError("ill-conditioned sequence: {}".format(seq))
    
    else:
        raise RuntimeError("illegal argument: {}".format(seq))

    return complex(seq[0],seq[2])
    
#---------------------------------------------------------------------------
def _simple_variance(v):
    v11, v12, v21, v22 = v
    if (
        abs(v11-v22) > 1E-15
    or  abs(v12) > 1E-15
    or  abs(v21) > 1E-15
    ):
        raise RuntimeError(
            "equal diagonal variance required, got: {}".format(v)
        )
#---------------------------------------------------------------------------
def mul2(arg1,arg2,estimated=False):
    """
    Return the product of ``arg1`` and ``arg2``

    Extends the usual calculation of a product, by
    using second-order contributions to uncertainty.
    
    :arg arg1: uncertain real or complex number
    :arg arg2: uncertain real or complex number
    :arg estimated: Boolean

    When both arguments are uncertain numbers 
    that have the same fixed values then 
    ``estimated`` should be set ``False``. 
    For instance, residual errors are often associated 
    with the value 0, or 1, which is not measured, in
    that case ``estimated=False`` is appropriate. 
    However, if either or both arguments are based on 
    measured values set ``estimated=True``.
    
    .. note::
    
        When ``estimated`` is ``True``, and the 
        product is close to zero, the result of a  
        second-order uncertainty calculation is 
        smaller than the uncertainty calculated by 
        the usual first-order method. In some cases, 
        an uncertainty of zero will be obtained.
    
    There are fairly strict limitations on the use of this
    function, especially for uncertain complex numbers:
    
    1) Arguments must be independent (have no common influence  
    quantities) and there can be no correlation between any 
    of the quantities that influence `arg1` or `arg2`. 

    2) If either argument is uncertain complex, the real and 
    imaginary components of that argument must have equal uncertainties  
    (i.e., the covariance matrix must be diagonal with equal elements 
    along the diagonal) and be independent (no common influences).

    A :class:`RuntimeError` exception is raised if  
    these conditions are not met.

    .. note::
    
        This function has been developed to improve the
        accuracy of uncertainty calculations where one or  
        both multiplicands are zero. In such cases, the 
        usual method of uncertainty propagation fails.

        For example ::
                
            >>> x1 = ureal(0,1,label='x1')
            >>> x2 = ureal(0,1,label='x2')
            >>> y = x1 * x2
            >>> y
            ureal(0.0,0.0,inf)
            >>> for cpt in rp.budget(y,trim=0):
            ... 	print("  {0.label}: {0.u}".format(cpt) )
            ... 	
              x1: 0.0
              x2: 0.0
              
        we see that none of the uncertainty in ``x1`` or ``x2`` 
        is propagated to ``y``. However, we may calculate 
        the second-order contribution ::
        
            >>> y = fn.mul2(x1,x2)
            >>> y
            ureal(0.0,1.0,inf)
            >>> for cpt in rp.budget(y,trim=0):
            ... 	print("  {0.label}: {0.u}".format(cpt) )
            ... 	
              x1: 0.70710678...
              x2: 0.70710678...
    
        The product now has a standard uncertainty of unity.
        
    .. warning::
    
        :func:`mul2` departs from the first-order linear  
        calculation of uncertainty described in the GUM.

        In particular, the strict proportionality between 
        components of uncertainty and first-order partial
        derivatives no longer holds.
        
        As a consequence, the calculation of first-order partial
        derivatives using the ``sensitivity`` method are also 
        incorrect.
        
    """
    reals = []
    comp = []
    for arg in (arg1,arg2):
        if not isinstance(arg,(UncertainReal,UncertainComplex)):
            raise RuntimeError(
                "uncertain number required, got: {!r}".format( arg )
            )

    if isinstance(arg1,UncertainReal):
        if isinstance(arg2,UncertainReal):
            return mult_2nd_real_pair(arg1,arg2,estimated)
        elif isinstance(arg2,UncertainComplex):
            _simple_variance(arg2.v)
            return mult_2nd_real_complex(arg1,arg2,estimated)
        else:
            raise RuntimeError(
                "uncertain number required, got: {!r}".format( arg2 )
            )
    elif isinstance(arg1,UncertainComplex):
        _simple_variance(arg1.v)
        if isinstance(arg2,UncertainReal):
            return mult_2nd_real_complex(arg2,arg1,estimated)
        elif isinstance(arg2,UncertainComplex):
            _simple_variance(arg2.v)
            return mult_2nd_complex_pair(arg1,arg2,estimated)
        else:
            raise RuntimeError(
                "uncertain number required, got: {!r}".format( arg2 )
            )
    else:
        raise RuntimeError(
            "uncertain number required, got: {!r}".format( arg1 )
        )
 
#---------------------------------------------------------------------------
def implicit(fn,x_min,x_max,epsilon=1E-13):
    """Return the solution to :math:`fn(x) = 0` 
        
    :arg fn: a user-defined function of one argument
    
    :arg x_min: lower limit of search range
    :type x_min: float

    :arg x_max: upper limit of search range     
    :type x_max: float

    :arg epsilon: tolerance for algorithm convergence   
    :type epsilon: float

    ``x_min`` and ``x_max`` delimit a range containing a single root 
    (ie, the function must cross the x-axis just once inside the range).

    .. note::
    
        *   A :class:`RuntimeError` is raised if the search algorithm fails to converge.
        
        *   An :class:`AssertionError` is raised if preconditions are not satisfied.

    **Example**::
    
        >>> near_unity = ureal(1,0.05)
        >>> fn = lambda x: x**2 - near_unity
        >>> function.implicit(fn,0,2)
        ureal(1.0,0.025...,inf)

    .. versionadded:: 1.3.4
    
    """
    return implicit_real(fn,x_min,x_max,epsilon)
    
#------------------------------------------------------------------------
def implicit_real(fn,x_min,x_max,epsilon):
    """Return the uncertain real number ``x``, that solves :math:`fn(x) = 0`

    The function fn() must take a single argument and x_min and
    x_max must define a range in which there is one (and only one)
    sign change in fn(x).

    The number 'epsilon' is a tolerance for accepting convergence.
    
    A RuntimeError will be raised if the root-search algorithm fails.

    An AssertionError will be raised if the preconditions for a 
    search to begin are not satisfied.

    Parameters
    ----------
    fn : a function of one argument
    x_min, x_max, epsilon : float

    Returns
    -------
    UncertainReal

    .. versionadded:: 1.3.4

    """
    xk,dy_dx = nr_get_root(fn,x_min,x_max,epsilon)

    # In an implicit function F(x,...) = 0, where
    # we solve F = 0 by finding a value for `x`,
    # `x` depends implicitly on the other arguments. 
    # The influence set of `F` and the influence set 
    # of `x` should are the same. 
    # `x` is not an elementary uncertain number;  
    # it is implicitly a function of the other 
    # arguments to F().

    # The components of uncertainty of `x` are related to 
    # the components of `F` as follows:
    #       u_i(x) = -( dF/dx_i / dF/dx ) * u_i(xi)

    y = fn( UncertainReal._constant(xk) )
    dx_dy = -1/dy_dx
    
    return UncertainReal(
        xk
    ,   vector.scale_vector(y._u_components,dx_dy)
    ,   vector.scale_vector(y._d_components,dx_dy)
    ,   vector.scale_vector(y._i_components,dx_dy)
    )     
    
#---------------------------------------------------------
# Newton-Raphson method with bisection.
# Based on Numerical Recipes in C, Ch 9, Section 4,
# but using uncertain numbers to evaluate the derivatives.
#
def nr_get_root(fn,x_min,x_max,epsilon):
    """Return the x-location of the root and the derivative at that point.
    
    This is a utility function used by implicit_real(). 
    It searches within the range for a real root.
    
    Parameters
    ----------
    fn : a function with a single argument
    x_min, x_max, epsilon : float

    Returns
    -------
    (float,float)

    .. versionadded:: 1.3.4

    """
    if x_max <= x_min:      
        raise RuntimeError(
            "Invalid search range: {!s}".format((x_min,x_max))
        )
           
    
    lower, upper = x_min, x_max

    ureal = lambda x,u: UncertainReal._elementary(x,u,inf,None,True)
    value = lambda x: x.x if isinstance(x,UncertainReal) else float(x)
    
    x = ureal(lower,1.0)
    f_x = fn(x) 
    fl = value(f_x)
    
    assert isinstance(f_x,UncertainReal),\
           "fn() must return an UncertainReal, got: %s" % type(f_x)
    
    if abs( fl ) < epsilon:
        return fl,f_x.sensitivity(x)
    
    x = ureal(upper,1.0)
    f_x = fn(x) 
    fu = value(f_x)

    if abs( fu ) < epsilon:
            return fu,f_x.sensitivity(x)

    if fl * fu >= 0.0:
        raise RuntimeError(
           "range does not appear to contain a root: {}".format((fl,fu))
        )

    if fl > 0.0:
        lower, upper = upper, lower

    # First place to look is the middle
    xk = (lower+upper) / 2.0  
    dx2 = abs(upper - lower)
    dx = dx2

    x = ureal(xk,1.0)
    f_x = fn(x) 
    f = value( f_x )
    df = f_x.sensitivity(x)

    for i in xrange(100): 
        if  (((xk-upper) * df-f) * ((xk-lower) * df - f) > 0.0
            or  ( abs(2.0 * f) > abs(dx2 * df) )):
            # Bisect if Newton out of range or
            # not decreasing fast enough.
            dx2 = dx
            dx = (upper - lower) / 2.0

            # If the bisection is too small then the root is found
            if(abs(dx) <= epsilon):
                return xk,df
            else:
                xk = lower + dx

        else:
            # Use Newton step
            dx2 = dx
            dx = f / df

            # If the change is ~ 0 then accept the root
            if(abs(dx) <= epsilon):
                return xk,df
            else:
                xk -= dx

        # Test convergence
        if(abs(dx) <= epsilon):
            return xk,df
  
        # Evaluate for next iteration;
        x = ureal(xk,1.0)
        f_x = fn(x) 
        f = value( f_x )
        df = f_x.sensitivity(x)

        if(f < 0.0):
            lower = xk
        else:
            upper = xk
            
    raise RuntimeError("Failed to converge") 
# ===========================================================================    
if __name__ == "__main__":
    import doctest
    from GTC import *
    doctest.testmod(  optionflags= doctest.NORMALIZE_WHITESPACE | doctest.ELLIPSIS  )


"""
Classes
-------
    * :class:`.UncertainArray`

Arithmetic operations
---------------------
    Arithmetic operations are defined for arrays
    (unary ``+`` and ``-``, and binary ``+``, ``-`` and ``*``).
    The multiplication  operator ``*`` is implemented element-wise.
    For two-dimensional arrays, matrix multiplication is performed
    by :func:`.matmul` (since Python 3.5, the ``@`` operator can be used).
    Also, :func:`.dot` evaluates the array dot product, which for
    two-dimensional arrays is equivalent to matrix multiplication.

    When one argument is a scalar, it is applied to each element
    of the array in turn.

Mathematical operations
-----------------------

    The standard mathematical operations defined in :mod:`.core`
    can be applied directly to an :class:`.UncertainArray`. An
    :class:`.UncertainArray` is returned, containing the result
    of the function applied to each element.

Functions
---------
    The functions :func:`inv`, :func:`transpose`, :func:`solve`
    and :func:`det` implement the usual linear algebra operations.

    The functions :func:`identity`, :func:`empty`, :func:`zeros`
    :func:`full` and :func:`ones` create simple arrays.

Reporting functions
-------------------

    Reporting functions :func:`~.reporting.u_component` and
    :func:`~.reporting.sensitivity` can be applied directly to
    a pair of arrays. An :class:`.UncertainArray` containing
    the result of applying the function to pairs of elements
    will be returned.

    The core `GTC` function :func:`~.core.result` can be used to
    define elements of an array as intermediate uncertain numbers.

Array broadcasting
------------------
    When binary arithmetic operations are applied to arrays, the shape
    of the array may be changed for the purposes of the
    calculation. The rules are as follows:

    *   If arrays do not have the same number of dimensions, then
        dimensions of size `1` are prepended to the smaller array's
        shape

    Following this, the size of array dimensions are compared and
    checked for compatibility. Array dimensions are compatible when

    *   dimension sizes are equal, or
    *   one of the dimension sizes is `1`

    Finally, if either of the compared dimension sizes is `1`, the
    size of the larger dimension is used. For example::

        >>> x = la.uarray([1,2])
        >>> y = la.uarray([[1],[2]])
        >>> print(x.shape,y.shape)  # doctest: +SKIP
        (2,) (2, 1)
        >>> x + y
        uarray([[2, 3],
                [3, 4]])

Module contents
---------------

"""
try:
    from itertools import izip  # Python 2
except ImportError:
    izip = zip

import numpy as np

from GTC import (
    LU,
    is_sequence,
)

__all__ = (
    'uarray',
    'dot',
    'matmul',
    'solve',
    'inv',
    'det',
    'identity','zeros','ones','empty','full',
    'transpose'
)

def uarray(array, label=None, names=None):
    """Create an array of uncertain numbers.

    For an overview on how to use an :class:`.UncertainArray` see :ref:`numpy-uarray`.

    .. versionadded:: 1.1

    .. attention::

       Requires numpy :math:`\geq` v1.13.0 to be installed.

    :param array: An array-like object containing :class:`int`, :class:`float`, :class:`complex`
                  :class:`~lib.UncertainReal` or :class:`~lib.UncertainComplex` elements.
    :param label: A label to assign to the `array`. This `label` does not
                  change labels previously assigned to array elements.
    :type label: str
    :param names: The field `names` to use to create a
                  :ref:`structured array <structured_arrays>`.
    :type names: list[str]

    :return: An :class:`.UncertainArray`.

    **Examples**:

        Create an `amps` and a `volts` array and then calculate the `resistances`

        >>> amps = la.uarray([ureal(0.57, 0.18), ureal(0.45, 0.12), ureal(0.68, 0.19)])
        >>> volts = la.uarray([ureal(10.3, 1.3), ureal(9.5, 0.8), ureal(12.6, 1.9)])
        >>> resistances = volts / amps
        >>> resistances
        uarray([ureal(18.070175438596493,6.145264246839438,inf),
                ureal(21.11111111111111,5.903661880050747,inf),
                ureal(18.52941176470588,5.883187720636909,inf)])

        Create a :ref:`Structured array <structured_arrays>`, with the names ``'amps'`` and ``'volts'``,
        and then calculate the `resistances`.

        >>> data = la.uarray([(ureal(0.57, 0.18), ureal(10.3, 1.3)),
        ...                (ureal(0.45, 0.12), ureal(9.5, 0.8)),
        ...                (ureal(0.68, 0.19), ureal(12.6, 1.9))], names=['amps', 'volts'])
        >>> resistances = data['volts'] / data['amps']
        >>> resistances
        uarray([ureal(18.070175438596493,6.145264246839438,inf),
                ureal(21.11111111111111,5.903661880050747,inf),
                ureal(18.52941176470588,5.883187720636909,inf)])

    """
    if np.__version__ < '1.13.0':
        # the __array_ufunc__ method was not introduced until version 1.13.0
        raise ValueError('creating an UncertainArray requires numpy >= 1.13.0')

    # don't allow a scalar UncertainArray
    if not (is_sequence(array) or (isinstance(array, np.ndarray) and array.ndim > 0)):
        raise ValueError('cannot create an UncertainArray from a scalar')

    dtype = None

    if names is not None:
        try:
            a_len = len(array[0])
            values = array[0]
        except:
            try:
                a_len = len(array)
                values = array
            except:
                a_len = None
                values = None

        if a_len is None or not isinstance(values, tuple):
            raise TypeError('The elements in the uarray must be a tuple if specifying field names')

        if a_len != len(names):
            raise ValueError('len(array[0]) != len(names) -> {} != {}'.format(a_len, len(names)))

        dtype = [(name, type(val)) for name, val in izip(names, array[0])]

    return UncertainArray(array, dtype=dtype, label=label)


def matmul(lhs, rhs):
    """Matrix product of a pair of two-dimensional arrays.

    For more details see :data:`numpy.matmul`.

    .. versionadded:: 1.1

    :param lhs: 2D array-like object.
    :param rhs: 2D array-like object.
    :return: The matrix product.
    :rtype: :class:`.UncertainArray`

    """
    # Must implement matrix multiplication because np.matmul does not
    # support dtype=object arrays in versions <= 1.15.4. Support for
    # np.matmul was added in numpy 1.16.0 as a ufunc.

    if not isinstance(lhs, np.ndarray):
        lhs = np.asarray(lhs)
    if not isinstance(rhs, np.ndarray):
        rhs = np.asarray(rhs)

    nd1, nd2 = lhs.ndim, rhs.ndim
    if nd1 == 0 or nd2 == 0:
        raise ValueError("Scalar operands are not allowed, use '*' instead")

    if nd1 <= 2 and nd2 <= 2:
        return UncertainArray(lhs.dot(rhs))

    broadcast = np.broadcast(np.empty(lhs.shape[:-2]), np.empty(rhs.shape[:-2]))
    ranges = [np.arange(s) for s in broadcast.shape]
    grid = np.meshgrid(*ranges, sparse=False, indexing='ij')
    indices = np.array([item.ravel() for item in grid]).transpose()

    i1 = indices.copy()
    i2 = indices.copy()
    for i in range(len(indices[0])):
        i1[:, i] = indices[:, i].clip(max=lhs.shape[i] - 1)
        i2[:, i] = indices[:, i].clip(max=rhs.shape[i] - 1)

    slices = np.array([[slice(None), slice(None)]]).repeat(len(indices), axis=0)
    i1 = np.hstack((i1, slices))
    i2 = np.hstack((i2, slices))
    out = np.array([matmul(lhs[tuple(a)], rhs[tuple(b)]) for a, b in izip(i1, i2)])
    return UncertainArray(out.reshape(list(broadcast.shape) + [lhs.shape[-2], rhs.shape[-1]]))


def dot(lhs, rhs):
    """Dot product of two arrays.

    For more details see :func:`numpy.dot`.

    .. versionadded:: 1.1

    :param lhs: The array-like object on the left-hand side.
    :param rhs: The array-like object on the right-hand side.
    :return: The dot product.
    :rtype: :class:`.UncertainArray`
    """
    return UncertainArray(np.dot(lhs, rhs))

def transpose(a, axes=None):
    """Array transpose

    For more details see :func:`numpy.transpose`.

    .. versionadded:: 1.1

    :param a: The array-like object
    :return: The transpose
    :rtype: :class:`.UncertainArray`

    """
    return UncertainArray(np.transpose(a, axes))

#---------------------------------------------------------------------------
def solve(a,b):
    """Return :math:`x`, the solution of :math:`a \cdot x = b`

    .. versionadded:: 1.1

    :arg a: 2D :class:`~uncertain_array.UncertainArray`
    :arg b: :class:`~uncertain_array.UncertainArray`

    :rtype: :class:`~uncertain_array.UncertainArray`


    **Example**::

        >>> a = la.uarray([[-2,3],[-4,1]])
        >>> b = la.uarray([4,-2])
        >>> la.solve(a,b)
        uarray([1.0, 2.0])

    """
    return LU.solve(a,b)


#---------------------------------------------------------------------------
def inv(a):
    """Return the (multiplicative) matrix inverse

    .. versionadded:: 1.1

    **Example**::

        >>> x = la.uarray( [[2,1],[3,4]])
        >>> x_inv =la.inv(x)
        >>> la.matmul(x,x_inv)
        uarray([[1.0, 0.0],
                [4.440892098500626e-16, 1.0]])

    """
    b = np.identity(a.shape[0],a.dtype)
    return LU.invab(a,b)

#---------------------------------------------------------------------------
def det(a):
    """Return the matrix determinant

    .. versionadded:: 1.1

    **Example**::

        >>> x = la.uarray( range(4) )
        >>> x.shape = 2,2
        >>> print(x)
        [[0 1]
        [2 3]]
        >>> la.det(x)
        -2.0

    """
    a_lu, i, p = LU.ludcmp(a.copy())
    return LU.ludet(a_lu,p)

#---------------------------------------------------------------------------
def identity(n):
    """Return an identity array with ``n`` dimensions

    .. versionadded:: 1.1

    **Example**::

        >>> la.identity(3)
        uarray([[1, 0, 0],
                [0, 1, 0],
                [0, 0, 1]])

    """
    return uarray( np.identity(n,dtype=object) )

def empty(shape):
    """Return an array of shape ``shape`` containing ``None`` elements

    .. versionadded:: 1.1

    **Example**::

        >>> la.empty( (2,3) )
        uarray([[None, None, None],
                [None, None, None]])

    """
    return uarray( np.empty(shape,dtype=object) )


def zeros(shape):
    """Return an array of shape ``shape`` containing ``0`` elements

    .. versionadded:: 1.1

    **Example**::

        >>> la.zeros( (2,3) )
        uarray([[0, 0, 0],
                [0, 0, 0]])

    """
    return uarray( np.zeros(shape,dtype=object) )

def ones(shape):
    """Return an array of shape ``shape`` containing ``1`` elements

    .. versionadded:: 1.1

    **Example**::

        >>> la.ones( (2,3) )
        uarray([[1, 1, 1], [1, 1, 1]])

    """
    return uarray( np.ones(shape,dtype=object) )

def full(shape,fill_value):
    """Return an array of shape ``shape`` containing ``fill_value`` elements

    .. versionadded:: 1.1

    **Example**::

        >>> la.full( (1,3),ureal(2,1) )
        uarray([[ureal(2.0,1.0,inf), ureal(2.0,1.0,inf),
                 ureal(2.0,1.0,inf)]])

    """
    return uarray( np.full(shape,fill_value,dtype=object) )


# import here to avoid circular imports when importing the matmul function in uncertain_array.py
from GTC.uncertain_array import UncertainArray

#============================================================================
if __name__ == "__main__":
    import doctest
    from GTC import *
    doctest.testmod(  optionflags=doctest.NORMALIZE_WHITESPACE )


import unittest
import copy
try:
    from itertools import izip  # Python 2
except ImportError:
    izip = zip

TOL = 1E-13
DIGITS = 13

from GTC.vector import (
    Vector,
    scale_vector,
    extend_vector,
    merge_weighted_vectors,
    merge_vectors,
    is_ordered
)

from testing_tools import *

#----------------------------------------------------------------------------
# The index of a vector is an object that must have a `uid` attribute
#
class Dummy(object):
    def __init__(self,uid):
        self.uid = (uid, 0)
        
#----------------------------------------------------------------------------
class TestVector(unittest.TestCase):
           
    # Vector objects are a core component. The speed of execution is 
    # critical to GTC. 
    
    # A Vector contains index-object reference pairs, where
    # the index values can be compared for order.
    
    # The implementation does not enforce ordering.
    # There is a utility function `is_ordered` available 
    # to check that Vector elements are correctly ordered
    
    def test_construct(self):
        """
        A Vector can be constructed empty and then updated
        or it can be constructed with named sequences, or it
        can be copy constructed.
        
        """
        v = Vector()
        self.assertEqual(len(v),0)
        
        uids = (1,2,3)
        indices = [ Dummy(i) for i in uids ]
        values = [12,34,45]
        
        v.extend(zip(indices,values))
        
        d = dict(zip(indices,values))
        for i,x in v.iteritems():
            self.assertTrue( equivalent(x,d[i]) )

        v = Vector(index=indices,value=values)
        for i,x in v.iteritems():
            self.assertEqual(x,d[i])

        vc = Vector(copy=v)
        for i,x in v.iteritems():
            self.assertEqual(x,d[i])
        
    def test_append(self):
        """
        Elements can be appended 
        
        """
        v = Vector()
        lst = [( Dummy(1),2),( Dummy(3),2)] 
        v.append( *lst[0] )     # First pair
        self.assertEqual(len(v),1)
        
        for p in v.items():
            self.assertEqual(lst[0][0],p[0])
            self.assertEqual(lst[0][1],p[1])

        v.append( *lst[1] )     # 2nd pair
        self.assertEqual(len(v),2)
        for i,p in enumerate(v.items()):
            self.assertEqual(lst[i][0],p[0])
            self.assertEqual(lst[i][1],p[1])

        self.assertTrue(is_ordered(v))
        
        lst[0] = ( Dummy(5),6)
        self.assertNotEqual(lst[0][0],p[0])
        self.assertNotEqual(lst[0][1],p[1])

    def test_scale(self):
        """
        A new vector can be created by scaling the
        values of another vector.
        
        """
        indices = [ Dummy(i) for i in range(9) ]
        values = list(range(9))
        indices_cp = copy.copy(indices)
        values_cp = copy.copy(values)
        
        v1 = Vector( index=indices,value=list(range(9)) )
        self.assertTrue(is_ordered(v1))
        
        k = 5.
        v2 = scale_vector(v1,k)
        
        # Make sure the arguments do not change 
        self.assertTrue( indices == indices_cp )
        self.assertTrue( values == values_cp )
        
        for i,x in v2.iteritems():
            self.assertEqual(i.uid[0]*k,x)

        self.assertTrue(is_ordered(v2))

        for i1,i2 in izip(v1.iterkeys(),v2.iterkeys()):
            self.assertEqual(i1.uid,i2.uid)

    def test_extend(self):
        """
        Extension is the process of making sure that one Vector has all the indices 
        in another Vector, as well as its own. 
        The initial values of any new indices are zero.
        
        """
        index1 = [ Dummy(i) for i in range(9) ]
        v1 = Vector( index=index1,value=list(range(9)) )
        index2 = [ Dummy(i) for i in range(5,15) ]
        v2 = Vector( index=index2,value=list(range(10)) )
        self.assertTrue(is_ordered(v1))
        self.assertTrue(is_ordered(v2))
        
        v = extend_vector(v1,v2)  
        
        v_range = [ i.uid[0] for i in v.keys() ]
        self.assertEqual(v_range,list(range(15)))
        self.assertTrue(is_ordered(v))
        
        for i,x in v.iteritems():
            # The values from v1 should remain,
            # those from v2 should be zero.
            if i.uid[0] < 9:
                self.assertEqual(x,i.uid[0])
            else:
                self.assertEqual(0,x)

    def test_merge_weighted_vectors(self):
        # Merge two overlapping continuous ranges
        rng1 = list(range(10))
        index1 = [ Dummy(i) for i in rng1 ]
        v1 = Vector( index=index1,value=rng1 )
        self.assertTrue(is_ordered(v1))
        k1 = 3.
        
        idx1_cp = copy.copy(index1)
        v1_cp = copy.copy(rng1)
        
        index2 = [ Dummy(i) for i in range(5,15) ]
        v2 = Vector( index=index2,value=rng1 )   # Note that the values are the same as v1
        self.assertTrue(is_ordered(v2))
        k2 = -4.
 
        idx2_cp = copy.copy(index2)
 
        v = merge_weighted_vectors(v1,k1,v2,k2)
        
        self.assertTrue(is_ordered(v))
        
        self.assertTrue( rng1 == v1_cp )
        self.assertTrue( index1 == idx1_cp )
        self.assertTrue( index2 == idx2_cp )
        
        rng = list(range(15))
        rng_merged = [ i.uid[0] for i in v.keys() ]
        self.assertEqual(rng_merged,rng)
        
        for i,p in enumerate( v.items() ):
            if i < 5:
                self.assertEqual(p[1],k1*rng[i])
            elif i > 4 and i < 10:
                self.assertEqual(p[1],k1*rng[i] + k2*rng[i-5])
            elif i > 9 and i < 15:
                self.assertEqual(p[1],k2*rng[i-5])
            else:
                assert False

    def test_merge_vectors(self):
        # Merge two overlapping continuous ranges
        rng1 = list(range(10))
        index1 = [ Dummy(i) for i in rng1 ]
        v1 = Vector( index=index1,value=rng1 )
        self.assertTrue(is_ordered(v1))
        
        idx1_cp = copy.copy(index1)
        v1_cp = copy.copy(rng1)
        
        index2 = [ Dummy(i) for i in range(5,15) ]
        v2 = Vector( index=index2,value=rng1 )   # Note that the values are the same as v1
        self.assertTrue(is_ordered(v2))
 
        idx2_cp = copy.copy(index2)
 
        v = merge_vectors(v1,v2)
        
        self.assertTrue(is_ordered(v))
        
        self.assertTrue( rng1 == v1_cp )
        self.assertTrue( index1 == idx1_cp )
        self.assertTrue( index2 == idx2_cp )
        
        rng = list(range(15))
        rng_merged = [ i.uid[0] for i in v.keys() ]
        self.assertEqual(rng_merged,rng)
        
        for i,p in enumerate( v.items() ):
            if i < 5:
                self.assertEqual(p[1],rng[i])
            elif i > 4 and i < 10:
                self.assertEqual(p[1],rng[i] + rng[i-5])
            elif i > 9 and i < 15:
                self.assertEqual(p[1],rng[i-5])
            else:
                assert False
                
    def test_merge_elementary(self):
        v1 = Vector( index=[Dummy(1)], value=[1] )
        v2 = Vector( index=[Dummy(2)], value=[1] )
        v = merge_weighted_vectors(v1,2,v2,3)
        self.assertEqual( len(v), 2)
        self.assertTrue(is_ordered(v))


#=====================================================
if(__name__== '__main__'):

    unittest.main()    # Runs all test methods starting with 'test'


import unittest

from GTC import *
from GTC.lib import (
    UncertainReal,
    complex_ensemble
)

from testing_tools import *

TOL = 1E-13 

#-----------------------------------------------------
class TestWillinkHall(unittest.TestCase):

    def test_pathological_cases(self):
        x1 = ucomplex( 0, (0.96,-0.34,-0.34,0.27), 5 )
        x2 = ucomplex( 0, (0.51,0.33,0.33,0.31), 3 )
        x3 = ucomplex( 1, (0.45,0.28,0.28,1.65), 6 )

        # The measurement equation
        x = x1*x2*x3

        equivalent( dof(x1), 5, 1E-15)
        equivalent( dof(x2), 3, 1E-15)
        equivalent( dof(x3), 6, 1E-15)
        # The components of uncertainty will all be zero
        self.assertTrue( dof(x) is nan )
        
        # Illegal correlation
        v,i = multiple_ucomplex(
            [4.999+0j,0.019661+0j],
            [(0.0032,0.0),(0.0000095,0.0)],
            5
        )
        phi = ucomplex(1.04446j,(0.0,0.00075),5,independent=False)

        self.assertRaises(RuntimeError,set_correlation,0.86,v.real,phi.imag)
        
    def testDoF1(self):
        # Test case from Metrologia 2002,39,365
        # Here define variables with covariance matrices
        TOL = 1E-2
        
        x1 = ucomplex( 1, (0.96,-0.34,-0.34,0.27), 5 )
        x2 = ucomplex( 1, (0.51,0.33,0.33,0.31), 3 )
        x3 = ucomplex( 1, (0.45,0.28,0.28,1.65), 6 )

        # The measurement equation
        x = x1 + x2 + x3
        
        equivalent( dof(x1), 5, 1E-15)
        equivalent( dof(x2), 3, 1E-15)
        equivalent( dof(x3), 6, 1E-15)

        self.assertTrue( equivalent( dof(x), 11.34 , TOL ) )

    def testDoF2(self):
        
        # Here define variables with standard uncertainties
        # and correlation coefficients

        TOL = 1E-2
        q1 = ucomplex( 1, (0.9789,0.51962), 5,independent=False )
        q2 = ucomplex( 1, (0.71414,0.5567), 3,independent=False )
        q3 = ucomplex( 1, (0.67082,1.28452), 6,independent=False )

        set_correlation( -0.6678, q1 )
        set_correlation( 0.82994, q2 )
        set_correlation( 0.32494, q3 )

        x = q1 + q2 + q3;
        
        self.assertTrue( equivalent( dof(x), 11.34 , TOL ) )

    def testDoF3(self):
        # Here we use a mixture of float and complex
        # uncertain numbers and expect to get the
        # usual GUM result for Example H1.
        
        TOL = 1E-10

        d_1 = ucomplex(0.0,(5.8,0),24)
        d_2 = ureal(0.0,3.9,5)
        d_3 = ucomplex(0.0,(6.7,0.0),8)

        # Note values from the guide are only approximate
        d = d_1 + d_2 + d_3
        df = dof(d)
        self.assertTrue( equivalent(df,25.4472507774,TOL) )

        alpha_s = ureal(0.0000115,tb.uniform(0.000002))
        delta_alpha = ureal(0.0,tb.uniform(0.000001),50)
        
        theta_1 = ucomplex(-0.1,(0.2,0.0))
        theta_2 = ucomplex(0.0,(tb.arcsine(0.5),0.0))
        delta_theta = ucomplex(0.0,(tb.uniform(0.05),0.0),2)

        Ls = ucomplex(5e7,(25.0,0.0),18)        

        theta = theta_1 + theta_2 
        df = dof(theta)
        self.assertTrue( math.isinf(df) )

        x1 = Ls * delta_alpha * theta       
        x2 = Ls * alpha_s * delta_theta        
        y = Ls + d - (x1 + x2)
        df = dof(y)
        self.assertTrue( equivalent(df,16.7521475092,TOL) )
        
    # Do a few tests of the invalid calculation to try to
    # cover various limit cases within the algorithm
    def testDoF4(self):
        
        # Correlations should cause the calculation to abort

        x1 = ucomplex( 1, (0.96,-0.34,-0.34,0.27), 5 )
        x2 = ucomplex( 1, (0.51,0.33,0.33,0.31), 5 )
        x3 = ucomplex( 1, (0.45,0.28,0.28,1.65), 5 )
        x = x1 + x2 + x3
        
        self.assertRaises(RuntimeError,set_correlation,0.5,x1.real,x2.real)

        # If we declare that they are in an ensemble then its OK
        complex_ensemble( (x1,x2,x3), 5 )
        self.assertTrue( not math.isnan(dof(x)) )

    def testDoF5(self):
        # Correlations should cause the calculation to abort

        x1 = ucomplex( 1, (0.96,-0.34,-0.34,0.27), 5 )
        x2 = ucomplex( 1, (0.51,0.33,0.33,0.31), 5 )
        x3 = ucomplex( 1, (0.45,0.28,0.28,1.65), 5 )
        x = x1 + x2 + x3

        self.assertRaises(RuntimeError,set_correlation,0.5,x1.real,x3.imag)

        # If we declare that they are in an ensemble then its OK
        complex_ensemble( (x1,x2,x3), 5 )
        self.assertTrue( not math.isnan(dof(x)) )

    def testDoF6(self):
        # Correlations should cause the calculation to abort

        x1 = ucomplex( 1, (0.96,-0.34,-0.34,0.27), 5 )
        x2 = ucomplex( 1, (0.51,0.33,0.33,0.31), 5 )
        x3 = ucomplex( 1, (0.45,0.28,0.28,1.65), 5 )
        x = x1 + x2 + x3

        self.assertRaises(RuntimeError,set_correlation,0.5,x2.imag,x3.real)

        # If we declare that they are in an ensemble then its OK
        complex_ensemble( (x1,x2,x3), 5 )
        self.assertTrue( not math.isnan(dof(x)) )

    def testDoF7(self):
        # Correlations should cause the calculation to abort

        x1 = ucomplex( 1, (0.96,-0.34,-0.34,0.27), 5 )
        x2 = ucomplex( 1, (0.51,0.33,0.33,0.31), 5 )
        x3 = ucomplex( 1, (0.45,0.28,0.28,1.65), 5 )
        x = x1 + x2 + x3

        self.assertRaises(RuntimeError,set_correlation,0.5,x2.imag,x3.imag)

        # If we declare that they are in an ensemble then its OK
        complex_ensemble( (x1,x2,x3), 5 )
        self.assertTrue( not math.isnan(dof(x)) )

    def testDoF8(self):
        # Look at the numerical stability of the dof calculation
        
        TOL = 1E-14
        
        cv1 = (0.96,-0.34,-0.34,0.27)
        cv2 = (0.51,0.33,0.33,0.31)
        cv3 = (0.45,0.28,0.28,1.65)
        
        scale = 1E-14

        scale_cv = lambda cv: [scale*x for x in cv ]
        
        x1 = ucomplex( 1, scale_cv(cv1), 5 )
        x2 = ucomplex( 1, scale_cv(cv2), 3 )
        x3 = ucomplex( 1, scale_cv(cv3), 6 )
        x = x1 + x2 + x3

        self.assertTrue( equivalent( dof(x), 11.3409777904914 , TOL ) )
        
    def testDoF9(self):
        TOL = 1E-5
        
        # v = ucomplex(4.999+0j,(0.0032,0.0),5,independent=False)
        # i = ucomplex(0.019661+0j,(0.0000095,0.0),5,independent=False)
        # phi = ucomplex(1.04446j,(0.0,0.00075),5,independent=False)
        
        v,i,phi = multiple_ucomplex(
            [4.999+0j,0.019661+0j,1.04446j],
            [(0.0032,0.0),(0.0000095,0.0),(0.0,0.00075)],
            5
        )

        set_correlation(-0.36,v.real,i.real)
        set_correlation(0.86,v.real,phi.imag)
        set_correlation(-0.65,i.real,phi.imag)

        z = v * exp(phi)/ i

        equivalent( uncertainty(z.real),0.0699787279884,TOL)
        equivalent( uncertainty(z.imag),0.295716826846,TOL)

        equivalent( get_correlation(z),-0.591484610819,TOL)

        equivalent( dof(z),5,TOL)

    def testDoF10(self):
        TOL = 1E-5
        
        values = [4.999+0j,0.019661+0j,1.04446j]
        uncert = [(0.0032,0.0),(0.0000095,0.0),(0.0,0.00075)]
        v,i,phi = multiple_ucomplex(values,uncert,5,('v','i','phi'))

        set_correlation(-0.36,v.real,i.real)
        set_correlation(0.86,v.real,phi.imag)
        set_correlation(-0.65,i.real,phi.imag)

        z = v * exp(phi)/ i

        equivalent( uncertainty(z.real),0.0699787279884,TOL)
        equivalent( uncertainty(z.imag),0.295716826846,TOL)

        equivalent( get_correlation(z),-0.591484610819,TOL)

        equivalent( dof(z),5,TOL)
        
        self.assertEqual(v.label,'v')
        self.assertEqual(i.label,'i')
        self.assertEqual(phi.label,'phi')

    def test_correlated_infinite_dof(self):
        # the routine should not have a problem when
        # influences with infinite dof are correlated

        # This is the same calculation as `testDoF7`
        # but with infinite dof for x2 and x3
        
        x1 = ucomplex( 1, (0.96,-0.34,-0.34,0.27), 5 )
        x2 = ucomplex( 1, (0.51,0.33,0.33,0.31) )
        x3 = ucomplex( 1, (0.45,0.28,0.28,1.65) )
        x = x1 + x2 + x3
        set_correlation(0.5,x2.imag,x3.imag)
        self.assertTrue( dof(x) is not nan )
 
    def test_correlated_ensemble(self):
        # test when non-consecutive influences 
        # are present in an ensemble
        
        TOL = 1E-5
        
        # Dummy UNs are created between the usual values
        v = ucomplex(4.999+0j,(0.0032,0.0),5,independent=False)
        dum1 = ucomplex( 1,1 )
        i = ucomplex(0.019661+0j,(0.0000095,0.0),5,independent=False)
        dum2 = ucomplex( 1,1 )
        phi = ucomplex(1.04446j,(0.0,0.00075),5,independent=False)

        complex_ensemble( (v,i,phi), 5 )

        set_correlation(-0.36,v.real,i.real)
        set_correlation(0.86,v.real,phi.imag)
        set_correlation(-0.65,i.real,phi.imag)

        # The dummy values are included with zero weight,
        # so the DoF calculation should not change.
        z = (v + 0*dum1) * exp(phi)/ (i + 0.0*dum2)

        equivalent( uncertainty(z.real),0.0699787279884,TOL)
        equivalent( uncertainty(z.imag),0.295716826846,TOL)

        equivalent( get_correlation(z),-0.591484610819,TOL)

        self.assertTrue( nan is not dof(z) )
        equivalent( dof(z),5,TOL)

    def test_illegal_complex_ensemble(self):
    
        # Check for inconsistent numbers of DoF 
        x1 = ucomplex( 1, (0.96,-0.34,-0.34,0.27), 5 )
        x2 = ucomplex( 1, (0.51,0.33,0.33,0.31) )
        x3 = ucomplex( 1, (0.45,0.28,0.28,1.65), 3 )
                 
        # Check for non-elementary UNs 
        # NB, this assertion may be removed in production
        y2 = x1 + x2 
        self.assertRaises(
            AssertionError,complex_ensemble,[x1,y2],5
        )        
              
    def testDoFMixedTypes(self):
        TOL = 1E-5
        
        # i = ucomplex(0.019661+0j,(0.0000095,0.0),5,independent=False)
        # phi = ucomplex(1.04446j,(0.0,0.00075),5,independent=False)
        
        i,phi = multiple_ucomplex(
            [0.019661+0j,1.04446j],
            [(0.0000095,0.0),(0.0,0.00075)],
            5
        )
        v = ureal(4.999,0.0032,5,independent=False)

        z = v * exp(phi)/ i
        
        # These are the values when correlation is ignored
        equivalent( uncertainty(z.real),0.194117890168,TOL)
        equivalent( uncertainty(z.imag),0.200665630894,TOL)

        equivalent( get_correlation(z),0.05820381031584,TOL)

        equivalent( dof(z),8.23564845495,TOL)              
#-----------------------------------------------------
class WelchSatterthwaiteExtensions(unittest.TestCase):
 
    def test_with_complex(self):
        """
        The real and imaginary components of a complex
        influence may be correlated
        
        """
        df_x1, df_x2, df_z1 = 5,6,7
        r = 0.1
        z1 = ucomplex(1,[1,r,r,1],df_z1)
        
        x1 = ureal(1,1,df_x1,independent=False)
        x2 = ureal(1,1,df_x2,independent=False)

        m = magnitude(z1)
        
        y = m + x1 + x2
        
        v_z1 = component(y,z1.real)**2\
               + component(y,z1.imag)**2\
               + 2*component(y,z1.real)*r*component(y,z1.imag)
        
        den = sum([
                component(y,x1)**4/df_x1,
                component(y,x2)**4/df_x2,
                v_z1**2/df_z1,
            ])
        nu_eff = variance(y)**2 / den
        
        df = dof(y)
        
        self.assertTrue( equivalent(nu_eff,df) )

        # but if others are correlated it fails
        self.assertRaises(RuntimeError,set_correlation,r,x1,x2)
        
    def test_with_complex_2(self):
        """
        The real and imaginary components of a complex
        quantity should behave as real UNs
        
        """
        x_re, u_x_re = -0.0152, 0.0040
        x_im, u_x_im = -0.0071, 0.0040 
        df_x = inf
        
        y_re, u_y_re = 0.0, 0.00057
        y_im, u_y_im = 0.0, 0.00072
        df_y = 3.0
        
        xre = ureal(x_re,u_x_re,df_x)
        xim = ureal(x_im,u_x_im,df_x)
        
        yre = ureal(y_re,u_y_re,df_y)
        yim = ureal(y_im,u_y_im,df_y)
        
        x = ucomplex(complex(x_re,x_im), [u_x_re,u_x_im], df_x)
        y = ucomplex(complex(y_re,y_im), [u_y_re,u_y_im], df_y)

        z = x + y
        zre = xre + yre
        zim = xim + yim
        
        self.assertTrue( dof(zre) != 0 )
        self.assertTrue( dof(zim) != 0 )
        
        self.assertTrue( equivalent(dof(z.real),dof(zre) ) )
        self.assertTrue( equivalent(dof(z.imag),dof(zim) ) )

    def test_with_infinity_1(self):
        """
        Influences may be correlated when they have
        infinite degrees of freedom
        
        """
        df_x3 = 5        
        x1 = ureal(1,1,independent=False)
        x2 = ureal(1,1,independent=False)
        x3 = ureal(1,1,df_x3)
        
        set_correlation(0.1,x1,x2)
        
        y = x1 + x2 + x3

        den = component(y,x3)**4/df_x3
        
        df = dof(y)
        nu_eff = variance(y)**2 / den
        
        self.assertTrue( equivalent(nu_eff,df))

    def test_with_infinity_2(self):
        """
        Complex components may be correlated when they have
        infinite degrees of freedom
        
        """
        df_x1, df_x2 = 5,6,
        r = 0.1
        x1 = ureal(1,1,df_x1)
        x2 = ureal(1,1,df_x2)
        z1 = ucomplex(1,[1,r,r,1])

        m = magnitude(z1)
        
        y = x1 + x2 + m
        
        den = sum([
                component(y,x1)**4/df_x1,
                component(y,x2)**4/df_x2,
            ])
        nu_eff = variance(y)**2 / den
        df = dof(y)
        
        self.assertTrue( equivalent(nu_eff,df))
        
#-----------------------------------------------------
class TestDoF(unittest.TestCase):

    """
    Additional tests related to DoF and ensemble calculations
    """

    def test_single_complex(self):
        """
        The DoF of a single complex quantity is a 
        special case in the WS algorithm 
        """
        z = ucomplex( (1+1j), [1.0,0.5,0.5,1.0], 6.0 )
        z_mag_sq = mag_squared(z)
        
        self.assertTrue( isinstance(z_mag_sq,UncertainReal) )
        self.assertTrue( equivalent(z_mag_sq.df,6) )
        

#============================================================================
if(__name__== '__main__'):

    unittest.main()    # Runs all test methods starting with 'test'


import unittest
import warnings 

from GTC import *

from testing_tools import *

TOL = 1E-13 

#----------------------------------------------------------------------------
class TestAPIFunctions(unittest.TestCase):
    """
    Test some general functions:
        - component
        - constant
        - variance
        - ureal + ucomplex check argument types
        - ureal properties: real, imag 
    """
    def test_ureal_float_interface(self):
        x = ureal(1.0,1.0)
        self.assertTrue( equivalent(x.x,x.real.x) )
        self.assertTrue( equivalent(x.u,x.real.u) )
        self.assertTrue( x is x.real )
        
        self.assertTrue( equivalent(0,x.imag.x) )
        self.assertTrue( equivalent(0,x.imag.u) )
       
        y = x + ureal(2,2)
        self.assertTrue( equivalent(y.x,y.real.x) )
        self.assertTrue( equivalent(y.u,y.real.u) )
        self.assertTrue( y is y.real )

        self.assertTrue( equivalent(0,y.imag.x) )
        self.assertTrue( equivalent(0,y.imag.u) )
        
    def test_ureal_arg_typs(self):
        """
        `ureal` should not accept anything but number types
        in its numerical arguments
        
        """
        x = "s"
        u = 0.1
        nu = 6.5
        self.assertRaises(TypeError,ureal,x,u,nu)
        x = ureal(1,1)
        self.assertRaises(TypeError,ureal,x,u,nu)
        x = 1
        u = 'd'
        self.assertRaises(TypeError,ureal,x,u,nu)
        u = ureal(1,1)
        self.assertRaises(TypeError,ureal,x,u,nu)
        u = 1
        nu = 'd'
        self.assertRaises(TypeError,ureal,x,u,nu)
        nu = ureal(1,1)
        self.assertRaises(TypeError,ureal,x,u,nu)
        
    def test_ucomplex_arg_typs(self):
        """
        `ureal` should not accept anything but number types
        in its numerical arguments
        
        """
        x = "s"
        u = (0.1,0.2)
        nu = 6.5
        self.assertRaises(TypeError,ucomplex,x,u,nu)
        x = ucomplex(1,1)
        self.assertRaises(TypeError,ucomplex,x,u,nu)
        x = 1
        u = 'd'
        self.assertRaises(TypeError,ucomplex,x,u,nu)
        u = ucomplex(1,1)
        self.assertRaises(TypeError,ucomplex,x,u,nu)
        u = 1
        nu = 'd'
        self.assertRaises(TypeError,ucomplex,x,u,nu)
        nu = ucomplex(1,1)
        self.assertRaises(TypeError,ucomplex,x,u,nu)
        
    def test_component(self):
        """
        component() should return the absolute value of a real component of uncertainty
        it should return the u_bar value of a complex component of uncertainty
        and it should return 0 if either argument is just a number.
        
        """
        x1 = ureal(1,1)
        x2 = -x1
        c = component(x2,x1)
        self.assertNotEqual(c,reporting.u_component(x2,x1))
        self.assertEqual(c,abs(reporting.u_component(x2,x1)))

        z1 = ucomplex(1+1j,(2,3))
        z2 = -z1
        c = component(z2,z1)
        uc = reporting.u_component(z2,z1)
        self.assertEqual(c,reporting.u_bar(uc))

        self.assertEqual(component(z2,3),0)
        self.assertEqual(component(x2,3),0)

    def test_constant(self):
        """
        A constant has no uncertainty
        it is of the uncertain type associated with its value type
        it may be labelled
        
        """
        x = constant(3)
        self.assertEqual( label(x), None )
        self.assertEqual( value(x),3) 
        self.assertEqual( uncertainty(x),0)
        
        x = constant(3,label='x')
        self.assertEqual( label(x), 'x' )
        self.assertEqual( value(x),3) 
        self.assertEqual( uncertainty(x),0)

        z = constant(3+1j)
        self.assertEqual( value(z),3+1j) 
        self.assertEqual( uncertainty(z),(0,0) )
        self.assertEqual( label(z), None )

        z = constant(3+1j,label='z')
        self.assertEqual( value(z),3+1j) 
        self.assertEqual( uncertainty(z),(0,0) )
        self.assertEqual( label(z), 'z' )

    def test_variance(self):
        """
        variance returns the std variance of a real number,
        or the std cv of a complex number,
        or 0
        
        """
        x = ureal(1,2)
        self.assertEqual(variance(x),4)

        cv = (5,3,3,4)        
        z = ucomplex(1+1j,cv)
        cv_z = variance(z)

        for i in range(4):
            self.assertTrue( equivalent(cv[i],cv_z[i],TOL) )
            
        self.assertEqual(variance(4),0)

    def test_get_correlation(self):
        # Self correlation tests 
        x = ureal(1,5)
        y = ureal(1,6)
        self.assertTrue( equivalent(
            get_correlation(x,x),
            1.0
        ) )
        self.assertTrue( equivalent(
            get_correlation(+x,x),
            1.0
        ) )
        self.assertTrue( equivalent(
            get_correlation(+x,+x),
            1.0
        ) )
        self.assertTrue( equivalent(
            get_correlation(x+0*y,x+0*y),
            1.0
        ) )
        self.assertTrue( equivalent(
            get_correlation(x+y,x+y),
            1.0
        ) )
        
        # If the product of variances is zero but the covariance
        # is also zero, then return a zero correlation coefficient.
        #
        l = constant(1)
        r = ureal(1,1) * ureal(1,1)
        cc = get_correlation(l,r)
        self.assertTrue( equivalent(cc,0.0) )
        
        # Asking for correlation wrt a number gives zero
        #
        x = ureal(1,1)
        cc = get_correlation(x,1.5)
        self.assertTrue( equivalent(cc,0.0) )
 
        z = ucomplex(1,1)
        cc = get_correlation(z,1.5+6j)
        self.assertTrue( equivalent_sequence(cc,[0.0,0.0,0.0,0.0]) )

        cc = get_correlation(1.5+6j)
        self.assertTrue( equivalent(cc,0.0) )

        # You can have complex number (implied) with no imaginary
        # component. If that is the only argument to get_correlation
        # then return zero
        cc = get_correlation(1.5)
        self.assertTrue( equivalent(cc,0.0) )
     
        # uc = ucomplex(1 + 1j, 0.1)
        # ur = ureal(2, 0.2)
        # r = 0.3
        # self.assertRaises(TypeError,set_correlation,r,uc,ur)
        
        # uc = ucomplex(1 + 1j, 0.1,independent=False)
        # ur = ureal(2, 0.2,independent=False)
        # set_correlation(r,uc,ur)
        # rr = get_correlation(uc,ur)
        
        
    def test_result(self):
            # `result()` should create a clone of the uncertain 
            # number object passed as an argument
            x = ureal(1.5,1)
            y = ureal(2,2)
            z = x**y 
            z_ = result(z)
            
            self.assertTrue( z is not z_ )
            self.assertTrue( equivalent(z.x,z_.x) )
            self.assertTrue( equivalent(z.u,z_.u) )
            self.assertTrue( equivalent(z.df,z_.df) )
            self.assertEqual( len(z._u_components), len(z_._u_components)) 
            self.assertEqual( len(z._i_components)+1, len(z_._i_components))    # a reference to self is added
            self.assertTrue( equivalent_sequence(z._u_components.values(),z_._u_components.values()) )
            
            x = ucomplex(1.2+.4j,1)
            y = ucomplex(-.5+1.1j,1)
            z = x * y 
            z_ = result(z)
            
            self.assertTrue( z is not z_ )
            self.assertTrue( equivalent_complex(z.x,z_.x) )
            self.assertTrue( equivalent_sequence(z.u,z_.u) )
            self.assertTrue( equivalent(z.df,z_.df) )
            self.assertEqual( len(z.real._u_components), len(z_.real._u_components)) 
            self.assertEqual( len(z.imag._u_components), len(z_.imag._u_components)) 
            self.assertEqual( len(z.real._i_components)+1, len(z_.real._i_components)) 
            self.assertEqual( len(z.imag._i_components)+1, len(z_.imag._i_components)) 
            self.assertTrue( equivalent_sequence(z.real._u_components.values(),z_.real._u_components.values()) )
            self.assertTrue( equivalent_sequence(z.imag._u_components.values(),z_.imag._u_components.values()) )

            # When applied to an elementary uncertain number nothing will happen 
            # unless the `label` argument is used. In that case, if the elementary 
            # uncertain number label is already set a warning is issued, 
            # otherwise label is assigned to the uncertain number.
            x = ureal(1.5,1)
            xx = result(x)
            self.assertTrue( x is xx )
            
            xx = result(x,label = 'x')
            self.assertTrue( x is xx )
            self.assertEqual( x.label, 'x' )
            
            x = ureal(1.5,1,label='x')
            xx = result(x,label='x')
            self.assertTrue( x is xx )
            self.assertEqual( x.label, 'x' )
 
            # From warnings documentation 2.7.16
            # Note that with Python 3 there is `assertWarns`
            with warnings.catch_warnings(record=True) as w:            
                # Cause all warnings to always be triggered.
                warnings.simplefilter("always")
                
                # Trigger a warning.
                xx = result(x,'xx')

                self.assertTrue( len(w) == 1 )
                self.assertTrue( issubclass(w[-1].category, RuntimeWarning) )
                self.assertTrue( x is xx )
                self.assertEqual( x.label, 'x' )
                
#----------------------------------------------------------------------------
class TestMultipleUN(unittest.TestCase):

    """
    """
    
    def test_multiple_ureal_creation(self):
        values = [4.999, 0.019661]
        uncert = [0.0032, 0.0000095]
        # this raises a RuntimeError: invalid `label_seq`: 'abc'
        self.assertRaises(RuntimeError,multiple_ureal,values, uncert, 5, label_seq='abc')
    
    def test_multiple_ucomplex_creation(self):
        values = [4.999 + 0j, 0.019661 + 0j]
        uncert = [(0.0032, 0.0), (0.0000095, 0.0)]
        # this raises a RuntimeError: invalid `label_seq`: 'abc'
        self.assertRaises(RuntimeError,multiple_ureal,values, uncert, 5, label_seq='abc')
    
#----------------------------------------------------------------------------
class TestGetSetCorrelation(unittest.TestCase):

    """
    tests the extended forms of get_correlation and
    set_correlation, where uncertain complex arguments
    are given.
    """
    
    def test(self):
        x1 = ureal(1,1)
        z1 = ucomplex(1,(1,1),independent=False)
        z2 = ucomplex(1,(1,1),independent=False)
        r = (.1,.2,.3,.4)
        set_correlation(r,z1,z2)
        check_r = get_correlation(z1,z2)
        self.assertTrue( equivalent_sequence(r,check_r) )

        r_t = (.1,.3,.2,.4)
        check_r = get_correlation(z2,z1)
        self.assertTrue( equivalent_sequence(r_t,check_r) )

        self.assertRaises(TypeError,set_correlation,r,z1,x1)
        self.assertRaises(TypeError,set_correlation,r,x1,z1)

    def test_illegal_set_correlation(self):
        x1 = ureal(1,2,3,independent=False)
        z1 = ucomplex(1,.5,3,independent=False)
        x2 = ureal(1,2,3,independent=False)
        z2 = ucomplex(1,.5,3,independent=False)
        
        r = 0.5
        
        self.assertRaises(RuntimeError,set_correlation,r,x1,x2)
        self.assertTrue( set_correlation(r,z1) is None )
        
    def test_with_mixed_unumbers(self):
        x1 = ureal(1,2,independent=False)
        z1 = ucomplex(1,.5,independent=False)
        
        r1 = .1
        set_correlation(r1,z1.real,x1)
        check_r = get_correlation(z1,x1)
        self.assertTrue( equivalent_sequence([r1,0,0,0],check_r) )
        
        r2 = -.3
        set_correlation(r2,z1.imag,x1)
        check_r = get_correlation(z1,x1)
        self.assertTrue( equivalent_sequence([r1,0,r2,0],check_r) )
        
        check_r = get_correlation(x1,z1)
        self.assertTrue( equivalent_sequence([r1,r2,0,0],check_r) )

        check_r = get_correlation(1.0,z1)
        self.assertTrue( equivalent_sequence([0]*4,check_r) )

        check_r = get_correlation(1.0+7j)
        self.assertTrue( equivalent(0,check_r) )

        check_r = get_correlation(1.0+7j,z1)
        self.assertTrue( equivalent_sequence([0]*4,check_r) )

        check_r = get_correlation(z1,1.0+7j)
        self.assertTrue( equivalent_sequence([0]*4,check_r) )

        check_r = get_correlation(1.0+7j,x1)
        self.assertTrue( equivalent_sequence([0]*4,check_r) )

        check_r = get_correlation(x1,1.0+7j)
        self.assertTrue( equivalent_sequence([0]*4,check_r) )
      
        check_r = get_correlation(7,x1)
        self.assertTrue( equivalent(0,check_r) )

        check_r = get_correlation(x1,1.0)
        self.assertTrue( equivalent(0,check_r) )
        
#----------------------------------------------------------------------------
class TestGetCovariance(unittest.TestCase):

    """
    tests get_covariance
    """
    
    def cv(self,u1,u2,r):
        return (
            u1[0]*r[0]*u2[0],
            u1[0]*r[1]*u2[1],
            u1[1]*r[2]*u2[0],
            u1[1]*r[3]*u2[1],
        )
    def test_self(self):
        # Self covariance tests 
        
        x = ureal(1,2)
        y = ureal(1,3)
        
        self.assertTrue( equivalent(
            get_covariance(x,x),
            x.u*x.u
        ) )
        self.assertTrue( equivalent(
            get_covariance(+x,x),
            x.u*x.u
        ) )
        self.assertTrue( equivalent(
            get_covariance(+x,+x),
            x.u*x.u
        ) )
        self.assertTrue( equivalent(
            get_covariance(x+0*y,x+0*y),
            x.u*x.u
        ) )
        self.assertTrue( equivalent(
            get_covariance(x+y,x+y),
            (x+y).u**2 
        ) )
        
    def test(self):
        ux1 = [0.25,0]
        uz1 = [0.33,1.5]
        uz2 = [.2,.4]
        x1 = ureal(1,ux1[0],independent=False)
        z1 = ucomplex(1,uz1,independent=False)
        z2 = ucomplex(1,uz2,independent=False)
        r = (.1,.2,.3,.4)
        
        set_correlation(r,z1,z2)
        check_r = get_covariance(z1,z2)
        self.assertTrue( equivalent_sequence(self.cv(uz1,uz2,r),check_r) )

        r_t = (.1,.3,.2,.4)
        check_r = get_covariance(z2,z1)
        self.assertTrue( equivalent_sequence(self.cv(uz2,uz1,r_t),check_r) )

        rx = [-.3,0,.7,0]
        set_correlation(rx[0],z1.real,x1)
        set_correlation(rx[2],z1.imag,x1)
        check_r = get_covariance(z1,x1)
        self.assertTrue( equivalent_sequence(self.cv(uz1,ux1,rx),check_r) )

        check_r = get_covariance(x1,z1)
        rx_t = [-.3,.7,0,0]
        self.assertTrue( equivalent_sequence(self.cv(ux1,uz1,rx_t),check_r) )

    def test_with_mixed_unumbers(self):
        ux1 = [2,0]
        x1 = ureal(1,ux1[0],independent=False)
        uz1 = [.5,.5]
        z1 = ucomplex(1,uz1,independent=False)
        
        r1 = .1
        set_correlation(r1,z1.real,x1)
        check_r = get_covariance(z1,x1)
        self.assertTrue( equivalent_sequence(self.cv(uz1,ux1,[r1,0,0,0]),check_r) )
        
        r2 = -.3
        set_correlation(r2,z1.imag,x1)
        check_r = get_covariance(z1,x1)
        self.assertTrue( equivalent_sequence(self.cv(uz1,ux1,[r1,0,r2,0]),check_r) )
        
        check_r = get_covariance(x1,z1)
        self.assertTrue( equivalent_sequence(self.cv(ux1,uz1,[r1,r2,0,0]),check_r) )

        check_r = get_covariance(1.0,z1)
        self.assertTrue( equivalent_sequence([0]*4,check_r) )

        check_r = get_covariance(1.0+7j)
        self.assertTrue( equivalent(0,check_r) )

        check_r = get_covariance(1.0+7j,z1)
        self.assertTrue( equivalent_sequence([0]*4,check_r) )

        check_r = get_covariance(z1,1.0+7j)
        self.assertTrue( equivalent_sequence([0]*4,check_r) )

        check_r = get_covariance(1.0+7j,x1)
        self.assertTrue( equivalent_sequence([0]*4,check_r) )

        check_r = get_covariance(x1,1.0+7j)
        self.assertTrue( equivalent_sequence([0]*4,check_r) )
      
        check_r = get_covariance(7,x1)
        self.assertTrue( equivalent(0,check_r) )

        check_r = get_covariance(x1,1.0)
        self.assertTrue( equivalent(0,check_r) )
        
#----------------------------------------------------------------------------
class TestPolymorphicMaths(unittest.TestCase):
    """
    math functions will assume that an uncertain number argument is provided,
    but should then try to find a suitable math or cmath library function 
    instead, finally raising TypeError if nothing can be done
    """
    def test_unary(self):
        unary_fns = [
            ('cos', 1.0, 1.0+.5j)
        ,   ('sin', 1.0, 1.0+.5j)
        ,   ('tan', 1.0, 1.0+.5j)
        ,   ('acos', 0.5, 0.5+0.5j)
        ,   ('asin', 0.5, 0.5+0.5j)
        ,   ('atan', 0.5, 0.5+0.5j)
        ,   ('exp', -0.5, -0.5+0.5j)
        ,   ('log', 0.5, 0.5+0.5j)
        ,   ('log10', 0.5, 0.5+0.5j)
        ,   ('sqrt',  0.5, 0.5+0.5j)
        ,   ('sinh', 1.0, 1.0+.5j)
        ,   ('cosh', 1.0, 1.0+.5j)
        ,   ('tanh', 1.0, 1.0+.5j)
        ,   ('acosh', 1.5, 1.5+0.5j)
        ,   ('asinh', 1.5, 1.5+0.5j)
        ,   ('atanh', 0.5, 0.5+0.5j)
        ]
        # This is intended to test the selection of 
        # an appropriate library or exception, not numerical accuracy.
        for d in unary_fns:
            self.assertTrue( equivalent(
                eval("{}({})".format(d[0],d[1])),
                eval("math.{}({})".format(d[0],d[1]))
            ) )
            self.assertTrue( equivalent_complex(
                eval("{}({})".format(d[0],d[2])),
                eval("cmath.{}({})".format(d[0],d[2]))
            ) )
            eval("self.assertRaises(TypeError,{},'s')".format(d[0]))
            
    def test_special_cases(self):
        """
        Note magnitude, mag_squared and pow do not use libraries
        
        """
        # phase:
        x = 1.0+1j
        self.assertTrue(
            equivalent(cmath.phase(x), phase(x) )
        )
        self.assertRaises(TypeError,phase,'z')
        
        # atan2:
        y, x = 1.0, 1.0
        self.assertTrue(
            equivalent(math.atan2(y,x), atan2(y,x))
        )
        self.assertRaises(TypeError,atan2,y,'x')
        self.assertRaises(TypeError,atan2,'y',x)
        self.assertRaises(TypeError,atan2,.5+.5j,.2+.2j)
       
        
#============================================================================
if(__name__== '__main__'):

    unittest.main()    # Runs all test methods starting with 'test'


import unittest
import sys
import math
import cmath
import numpy
import operator
import collections
try:
    xrange  # Python 2
except NameError:
    xrange = range

TOL = 1E-13
DIGITS = 13

from GTC import *
from GTC.LU import *

from testing_tools import *

#-----------------------------------------------------------------------------------
def one_port_xratio(measured,nominal):
    """
    'nominal' and 'measured' are 3-element sequences for the three artefacts used
    to calibrate the one-port.

    returns the three elements: E_D, E_S, E_R

    Equations are from R. W. Beatty and A. C. Macpherson, proc. I.R.E. vol 41, (9) pp 1112-1119 (1953),
    also in R. W. Beatty, "Invariance of the cross ratio applied to microwave network analysis"
    (NBS TN-623, Sept. 1972)

    NB
    * The NBS Note has two transcription errors in the formulae.
    * This algorithm appears to be about 20% SLOWER than using linear algebra (with mst.GUM)
    * The results of this method and the LA method differ when problems are not well-conditioned
    
    """
    m_1, m_2, m_3 = measured
    n_1, n_2, n_3 = nominal

    n_12 = n_1 * n_2     
    n_23 = n_2 * n_3    
    n_31 = n_1 * n_3

    dm_12 = m_1 - m_2
    dm_23 = m_2 - m_3
    dm_31 = m_3 - m_1

    # Common demoninator
    den = n_12 * dm_12 + n_23 * dm_23 + n_31 * dm_31

    # Directivity
    num = n_12 * m_3 * dm_12 + n_23 * m_1 * dm_23 + n_31 * m_2 * dm_31
    E_D = num / den

    # Source match
    num = n_1 * dm_23 + n_2 * dm_31 + n_3 * dm_12
    E_S = -num / den

    # Reflection tracking
    num = dm_12 * dm_23 * dm_31 * (n_1 - n_2) * (n_2 - n_3) * (n_3 - n_1)
    E_R = num / den**2

    return dict(E_D=E_D,E_S=E_S,E_R=E_R)

#-----------------------------------------------------------------------------------
def one_port(measured,nominal):
    """
    -> a sequence containing E_D, E_S, E_R for a 1-port 

    'measured','nominal' are 3-element sequences
        
    """    
    # H =   [ (nominal[0], unity, -nominal[0] * measured[0]),
    #           etc
    #       ]
    H = numpy.array( [ (n,1.0,-n * m) for m,n in zip(measured,nominal) ] )
    b = numpy.array( measured )
    
    ABC = solve( H,b )

    E_D=ABC[1]
    E_S=-ABC[2]
    E_R=ABC[0] - ABC[1] * ABC[2]
    
    return dict(E_D=E_D,E_S=E_S,E_R=E_R)


#-----------------------------------------------------
class TestLUScalar(unittest.TestCase):
    """
    The LU module provides functions to:
        * solve a system of linear equations,
        * evaluate the determinant of a matrix,
    
    """
    def test(self):
        
        a = numpy.array(
            [ (17.99753493515198, 18.08102451776513, 6),
              (627.1639334786039, 27.20225830408064, 8.998767467575989),
              (27.20225830408064, 628.4433741012242, 9.040512258882565)
            ])

        b = numpy.array( (627.8036537899141, 1845.420819171415, 1851.64204122553) )

        # determinant evaluated using numpy.linalg.det(a)     
        numpy_det = 2164937.084414148 
        
        a_lu,i,p = ludcmp( a.copy() )
        lu_det = ludet(a_lu,p)
        self.assertTrue( equivalent(numpy_det,lu_det) )

        # Solve system of equations
        numpy_soln = numpy.linalg.solve( a, b )
        lu_soln = solve( a,b )
        for b1,b2 in zip(numpy_soln,lu_soln):
            self.assertTrue( equivalent(b1,b2) )

#-----------------------------------------------------
class TestLUInvProduct(unittest.TestCase):
    """
    The LU module provides a function to evaluate inv(a).b
    where `b` is a 2D array
    
    """
    def test(self):
        a = numpy.array(
            [ (17.99753493515198, 18.08102451776513, 6),
              (627.1639334786039, 27.20225830408064, 8.998767467575989),
              (27.20225830408064, 628.4433741012242, 9.040512258882565)
            ])

        # Calculate the inverse matrix
        a_inv = invab(a,numpy.identity(3))
        a_b = numpy.dot(a,a_inv)
        for i in xrange(3):
            for j in xrange(3):
                if i == j:
                    self.assertTrue( equivalent(1.0,a_b[i,i]) )   
                else:
                    self.assertTrue( equivalent(0.0,a_b[i,j]) )    

        # Solve `a.x=b` for `b` a matrix 
        x = numpy.array([ (-2,1), (-3,2), (4,4) ])
        b = numpy.dot(a,x)
        
        a_b = invab(a,b)
        for i in xrange(a_b.shape[0]):
            for j in xrange(a_b.shape[1]):
                self.assertTrue( equivalent(x[i,j],a_b[i,j]) )

#-----------------------------------------------------
class TestLUComplex(unittest.TestCase):
    """
    The LU module can be used to solve systems of
    linear equations involving complex quantities.
    
    """
    def test_one_port_x_ratio(self):
        """
        The Beatty cross-ratio method should give the same results
        
        """
        measured = (-0.188 - 0.902j, 0.239 + 0.936j, 0.006 + 0.007j)
        nominal = (-1 + 0j,1 + 0j,0 + 0j)

        result1 = one_port_xratio(measured,nominal)
        result2 = one_port(measured,nominal)

        for k in result1:
            equivalent_complex(result1[k],result2[k],TOL)

    def test_known_errors(self):
        """
        Create some 'measured' data by passing known 
        standard values through a known error network
        and check that the errors can be recovered.
        
        """
        error = dict(E_D = 0.04+0.01j,E_S=0.15-0.02j,E_R=0.97+0.01j)
        standards = dict(G_1=0.03,G_2=0.95+0.1j,G_3=-0.97-0.01j)
        
        # transforms gamma to one seen behind the error box 
        measure = lambda G: error['E_D'] + G*error['E_R']/(1-error['E_S']*G)
        
        nominal = standards.values()
        measured = [ measure(G) for G in nominal ]

        result2 = one_port(measured,nominal)
        for k in error:
            equivalent_complex(error[k],result2[k],TOL)

#-----------------------------------------------------
class TestLUUncertainComplex(unittest.TestCase):
    """
    This test uses analytical results by Rodriguez, for the
    one-port calibration case, to compare the sensitivity
    coefficients obtained after a one-port calibration.
    
    """
    def test(self):
        seq_to_complex = lambda seq: complex(seq[0],seq[2])

        #-------------------------------------------------------
        # The values of the standard artefacts 
        # Use the unity uncertainty for each standard
        # and construct a vector of uncertain number inputs.
        #
        Open = 0.98 + 0.01j
        Short = -0.95 -0.03j
        Load = 0.0 + 0.0j

        Gx = [Open, Short, Load]
        G = [ ucomplex(x,[1,1]) for x in Gx ]

        # The measured values 
        # These are just dreamed up
        Gm = [-0.188 - 0.902j, 0.239 + 0.936j, 0.006 + 0.007j ]
        T = [ ucomplex(x,[1,1]) for x in Gm ]

        p1 = one_port(T,G)
        E_S = p1['E_S']
        E_R = p1['E_R']
        E_D = p1['E_D']

        # Values to put in Rodriguez equations
        e_s = value(E_S)
        e_r = value(E_R)
        e_d = value(E_D)
        g_sc = value(G[0])
        g_oc = value(G[1])
        g_ld = value(G[2])
        a = value(T[2])
        b = value(T[1])
        c = value(T[0])

        ## =================== Rodriguez calculated equivalents ====================
        # Note that there is a systematic sign error in Rodriguez equation,
        # negative signs here correct Rodriguez sign error
        de_s_d_g_sc = -(1-e_s*g_oc) / (g_sc * (g_sc -g_oc)) # short
        de_s_d_g_oc = -(1-e_s*g_sc) / (g_oc * (g_oc - g_sc)) # open
        de_s_d_g_ld = -( (1 - e_s*g_oc)*(1-e_s*g_sc)/(g_oc*g_sc) ) #load

        de_s_d_C =(1-e_s*g_sc)**2 * (1-e_s*g_oc) / (g_sc * e_r *(g_sc -g_oc)) # short
        de_s_d_B = (1-e_s*g_oc)**2 * (1-e_s*g_sc) / (g_oc * e_r * (g_oc - g_sc)) # open
        de_s_d_A= ( (1 - e_s*g_oc)*(1-e_s*g_sc)/(g_oc*g_sc*e_r) ) #load

        ## ============ One port error terms and covariances for full solution =====================
        dE_S_d_g_sc = seq_to_complex( rp.u_component(E_S,G[0]) ) # short
        dE_S_d_g_oc = seq_to_complex( rp.u_component(E_S,G[1]) ) # open
        dE_S_d_g_ld = seq_to_complex( rp.u_component(E_S,G[2]) ) # load

        dE_S_d_C = seq_to_complex( rp.u_component(E_S,T[0]) )
        dE_S_d_B = seq_to_complex( rp.u_component(E_S,T[1]) )
        dE_S_d_A = seq_to_complex( rp.u_component(E_S,T[2]) )

        equivalent_complex(de_s_d_g_sc,dE_S_d_g_sc,TOL)
        equivalent_complex(de_s_d_g_oc,dE_S_d_g_oc,TOL)
        equivalent_complex(de_s_d_g_ld,dE_S_d_g_ld,TOL)

        equivalent_complex(de_s_d_C,dE_S_d_C,TOL)
        equivalent_complex(de_s_d_B,dE_S_d_B,TOL)
        equivalent_complex(de_s_d_A,dE_S_d_A,TOL)

        ## Compare error terms from full method with
        ## Rodriguez calculated equivalents 
        equivalent_complex( e_s, ((a-b)/g_oc-(a-c)/g_sc)/(c-b), TOL)
        equivalent_complex( e_r ,(a-b)*(a-c)*(1/g_oc-1/g_sc)/(c-b), TOL)
        equivalent_complex( e_d , a, TOL)
        
#=====================================================
if(__name__== '__main__'):

    unittest.main()    # Runs all test methods starting with 'test'